from django import forms
from .models import Entry, Customer
from datetimewidget.widgets import (
    TimeWidget,
    DateWidget,
)
from . import settings
from .widgets import RelatedFieldWidgetCanAdd



TIME_FORMATS = ['%H:%M', '%I:%M%p', '%I:%M %p',]
DURATION_FORMATS = ['%H:%M',]
DATE_WIDGET_OPTIONS = {
    # NB convert between conventions for numbering days (widget uses Sunday=0)
    'weekStart': ((settings.DIARY_FIRST_DAY_OF_WEEK + 1) % 7),
    'minView': 2,
    'maxView': 3,
}
TIME_WIDGET_OPTIONS = {
    'format': 'HH:ii P' if settings.DIARY_SHOW_MERIDIAN else 'hh:ii',
    'showMeridian': settings.DIARY_SHOW_MERIDIAN,
    'minuteStep': 15,
}
DURATION_WIDGET_OPTIONS = {
    'format': 'hh:ii',
    'minuteStep': 15,
}



class EntryForm(forms.ModelForm):


    def __init__(self, *args, **kwargs):
        """
        Override constructor to allow customer field to be configurably visible.
        """
        exclude_customer = kwargs.pop('exclude_customer', False)
        super(EntryForm, self).__init__(*args, **kwargs)
        if exclude_customer:
            del self.fields['customer']
        else:
            # make sure the customer selection popup has some sort of sane order
            self.fields['customer'].queryset = Customer.objects.all().order_by(
                'first_name',
                'last_name',
            )


    class Meta:
        model = Entry
        fields = (
            'customer',
            'date',
            'time',
            'treatment',
            'duration',
            'resource',
            'notes',
        )
        widgets = {
            # override customer widget attributes AFTER creating the form
            'customer': RelatedFieldWidgetCanAdd(Customer),
            'date': DateWidget(
                bootstrap_version=3,
                options=DATE_WIDGET_OPTIONS,
            ),
            'time': TimeWidget(
                bootstrap_version=3,
                options=TIME_WIDGET_OPTIONS,
            ),
            'duration': TimeWidget(
                bootstrap_version=3,
                options=DURATION_WIDGET_OPTIONS,
            ),
        }
        input_formats = {
            'time': TIME_FORMATS,
            'duration': DURATION_FORMATS,
        }



from django.test import TestCase
from django.test.utils import override_settings
from django.utils import timezone
import datetime
from django.forms import ValidationError
import traceback
from django.contrib.auth.models import User
from django.urls import reverse
from django.conf import settings as main_settings
from . import settings
from . import views
from .views import get_today_now
import importlib as imp         # since Python 3.4
from django.core.management import call_command
from django.core.management.base import CommandError
from freezegun import freeze_time


# Create your tests here.

from .models import Customer, Treatment, Resource, Entry
from django.contrib.auth.models import User


TIME_ZERO = datetime.time(0,0)


def yearsago(years, from_date=None):
    """
    Workaround for datetime.timedelta not knowing how to calculate leap years.
    Adapted from:
    http://stackoverflow.com/questions/765797/python-timedelta-in-years
    """
    if from_date is None:
        from_date = timezone.now()
    try:
        return from_date.replace(year=from_date.year - years)
    except:
        # Must be 2/29!
        assert from_date.month == 2 and from_date.day == 29 # can be removed
        return from_date.replace(month=2, day=28,
                                 year=from_date.year-years)



class CustomerTests(TestCase):


    def test_customer_age(self):
        """
        Make sure the customer age calculation works correctly and as expected.
        """
        tenYearsAgo = yearsago(10)
        tenYearCustomer = Customer.objects.create(
            username='Ten',
            password='random',
            date_of_birth=tenYearsAgo,
        )
        self.assertEqual(tenYearCustomer.age(), 10)


    def test_customer_age_date_of_birth_not_set(self):
        """
        Make sure something sensible happens when we calculate age from an
        undefined birth date.
        """
        undefinedBirthCustomer = Customer.objects.create(
            username='Ageless',
            password='random',
        )
        self.assertIsNone(undefinedBirthCustomer.age())



def create_customer(username):
    """
    Utility to generate a customer
    """
    customer = Customer.objects.create_user(
        username,                                   # username
        username,                                   # first_name
        username,                                   # last_name
        username+'@example.com',                    # email
        '123456',                                   # phone
        datetime.date(year=1950, month=6, day=1),   # d-o-b
        'M',                                        # gender
        'test user',                                # notes
    )
    customer.save()
    return customer


def obtain_superuser():
    """
    Utility to create a power user in the test database
    """
    superuser = User.objects.filter(is_superuser=True).first()
    if not superuser:
        superuser = User.objects.create_superuser(
            'admin',
            'admin@example.com',
            'password'
        )
        superuser.save()
    return superuser


def create_treatment(name, min_duration, resource_required):
    """
    Utility to create a treatment.
    """
    treatment = Treatment(
        name=name,
        min_duration=min_duration,
        resource_required=resource_required,
    )
    treatment.save()
    return treatment


def create_resource(name, description):
    """
    Utility to make a resource
    """
    resource = Resource(
        name=name,
        description=description,
    )
    resource.save()
    return resource


def change_time(date, time, delta):
    """
    Utility to change a time by a preset amount described by a delta.
    The date is just a tool.
    """
    return (datetime.datetime.combine(date, time) + delta).time()


def create_entry_from_delta(dateDelta, time, duration, notes):
    """
    Utility to create an entry for testing using a dateDelta.
    """
    date = timezone.datetime.today() + dateDelta
    return create_entry(date, time, duration, notes)


def create_entry(date, time, duration, notes):
    """
    Utility to create an entry for testing using a date.
    Updated to use timefield for duration.
    Updated to add a power user as default creator.
    Updated to add a power user as default editor.
    """
    duration_as_time = change_time(date, TIME_ZERO, duration)
    entry = Entry(
        notes=notes,
        date=date,
        time=time,
        duration=duration_as_time, # duration in the database is really a time
    )
    entry.creator = obtain_superuser()
    entry.editor = obtain_superuser()
    return entry


class EntryModelTests(TestCase):
    """
    Tests of the entry model methods.
    These tests are primarily about making sure entries can do basic time
    arithmetic and comparisons, with the intention that they can determine time
    collisions.
    """


    def test_time_end_calculation(self):
        """
        Make sure the time arithmetic works as expected.
        """
        dateDelta = datetime.timedelta(days=0)
        time = datetime.time(hour=12)
        duration = datetime.timedelta(hours=1)
        entry = create_entry_from_delta(
            dateDelta,
            time,
            duration,
            'time calc test 1',
        )
        result = datetime.time(hour=13)
        self.assertEqual(entry.time_end(), result)


    def test_entry_treatment_resource_required(self):
        """
        If the entry has a treatment that specifies a resource a resource must
        be defined.
        """
        entry = create_entry_from_delta(
            datetime.timedelta(days=0),
            datetime.time(hour=12),
            datetime.timedelta(hours=0),
            'resource_required_test',
        )
        # make sure an exception is NOT raised when there is no treatment
        try:
            entry.clean()
            self.assertFalse(entry is None)
        except Exception as e:
            traceback.print_exc()
            self.fail(
                'Cleaning an entry with no treatment '\
                'raised an unexpected exception: \n{0}'.format(e)
            )

        # add a treatment that needs a resource
        entry.treatment = create_treatment(
            'requires_resource',
            datetime.timedelta(hours=1),
            True,
        )
        self.assertRaisesMessage(
            ValidationError,
            'Resource requirement is not met.',
            entry.clean,
        )

        # now add a resource, get new exception due to duration
        entry.resource = create_resource(
            'resource',
            'resource',
        )
        self.assertRaisesMessage(
            ValidationError,
            'Duration must be at least the minimum treament time.',
            entry.clean,
        )

        # change the duration to the min required, clean is now ok
        entry.duration = datetime.time(hour=1)
        try:
            entry.clean()
            self.assertFalse(entry is None)
        except Exception as e:
            traceback.print_exc()
            self.fail(
                'Cleaning an entry with\n'\
                'treatment resource and duration '\
                'raised an unexpected exception: \n{0}'.format(e)
            )


    def test_entry_equality_different_days(self):
        """
        Entries on different days are not equal.
        """
        dateDelta1 = datetime.timedelta(days=0)
        time = datetime.time(hour=12)
        duration = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta1,
            time,
            duration,
            'time calc test 1',
        )

        dateDelta2 = datetime.timedelta(days=1)
        entry2 = create_entry_from_delta(
            dateDelta2,
            time,
            duration,
            'time calc test 1',
        )

        self.assertFalse(entry1 == entry2)


    def test_entry_equality_identity(self):
        """
        Entries with identical dates, times and durations are equal.
        """
        dateDelta1 = datetime.timedelta(days=0)
        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'time calc test 1',
        )

        dateDelta2 = datetime.timedelta(days=0)
        time2 = datetime.time(hour=12)
        duration2 = datetime.timedelta(hours=1)
        entry2 = create_entry_from_delta(
            dateDelta2,
            time2,
            duration2,
            'time calc test 1',
        )

        self.assertTrue(entry1 == entry2)


    def test_entry_equality_no_overlap_before(self):
        """
        Entries on the same day where the first ends before the second starts
        are not equal.
        """
        dateDelta = datetime.timedelta(days=0)

        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta,
            time1,
            duration1,
            'time calc test 1',
        )

        time2 = datetime.time(hour=14)
        duration2 = datetime.timedelta(hours=1)
        entry2 = create_entry_from_delta(
            dateDelta,
            time2,
            duration2,
            'time calc test 1',
        )

        self.assertFalse(entry1 == entry2)


    def test_entry_equality_no_overlap_after(self):
        """
        Entries on the same day where the first starts after the second ends are
        not equal.
        """
        dateDelta = datetime.timedelta(days=0)

        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta,
            time1,
            duration1,
            'time calc test 1',
        )

        time2 = datetime.time(hour=9)
        duration2 = datetime.timedelta(hours=1)
        entry2 = create_entry_from_delta(
            dateDelta,
            time2,
            duration2,
            'time calc test 1',
        )

        self.assertFalse(entry1 == entry2)


    def test_entry_equality_consecutive(self):
        """
        Entries on the same day where the first and second
        events are consecutive are not equal.
        """
        dateDelta = datetime.timedelta(days=0)

        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta,
            time1,
            duration1,
            'time calc test 1',
        )

        time2 = entry1.time_end()
        duration2 = datetime.timedelta(hours=1)
        entry2 = create_entry_from_delta(
            dateDelta,
            time2,
            duration2,
            'time calc test 1',
        )

        self.assertFalse(entry1 == entry2)


    def test_entry_equality_envelope(self):
        """
        Entries on the same day where one event encompasses the other are
        'equal' (i.e. they overlap in time).
        """
        dateDelta = datetime.timedelta(days=0)

        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=3)
        entry1 = create_entry_from_delta(
            dateDelta,
            time1,
            duration1,
            'time calc test 1',
        )

        time2 = datetime.time(hour=13)
        duration2 = datetime.timedelta(hours=1)
        entry2 = create_entry_from_delta(
            dateDelta,
            time2,
            duration2,
            'time calc test 1',
        )

        self.assertTrue(entry1 == entry2)


    def test_entry_resource_clash_clean_raises_exception(self):
        """
        Make sure cleaning the data raises the correct Exception when there is a
        resource clash.
        """
        resource = create_resource('resource', 'resource')

        dateDelta1 = datetime.timedelta(days=0)
        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'time calc test 1',
        )
        entry1.resource = resource
        entry1.save()

        dateDelta2 = datetime.timedelta(days=0)
        time2 = datetime.time(hour=12)
        duration2 = datetime.timedelta(hours=1)
        entry2 = create_entry_from_delta(
            dateDelta2,
            time2,
            duration2,
            'time calc test 1',
        )
        entry2.resource = resource

        self.assertTrue(entry1 == entry2)
        self.assertRaisesMessage(
            ValidationError,
            'Resource clash with another Entry. Please change resource or time.',
            entry2.clean,
        )


    def test_cancelled_entry_resource_clash_clean_no_exception(self):
        """
        Make sure cleaning the data raises no Exception when the entry is
        cancelled even with a resource clash.
        """
        resource = create_resource('resource', 'resource')

        dateDelta1 = datetime.timedelta(days=0)
        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'time calc test 1',
        )
        entry1.resource = resource
        entry1.save()

        dateDelta2 = datetime.timedelta(days=0)
        time2 = datetime.time(hour=12)
        duration2 = datetime.timedelta(hours=1)
        entry2 = create_entry_from_delta(
            dateDelta2,
            time2,
            duration2,
            'time calc test 1',
        )
        entry2.resource = resource
        entry2.cancelled = True

        # make sure an exception is NOT raised
        try:
            entry2.clean()
            self.assertTrue(entry1 == entry2)
        except Exception as e:
            traceback.print_exc()
            self.fail(
                'Cleaning a cancelled entry with \n'\
                'a time clash raised an unexpected exception: \n{0}'.format(e)
            )


    def test_entry_no_resource_clash_no_exception(self):
        """
        Make sure cleaning the data raises no Exception when resources don't
        clash with pre-existing entry data.
        """
        resource = create_resource('resource', 'resource')

        dateDelta1 = datetime.timedelta(days=0)
        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'time calc test 1',
        )
        entry1.resource = resource
        entry1.save()

        resource2 = create_resource('resource_2', 'resource_2')

        dateDelta2 = datetime.timedelta(days=0)
        time2 = datetime.time(hour=12)
        duration2 = datetime.timedelta(hours=1)
        entry2 = create_entry_from_delta(
            dateDelta2,
            time2,
            duration2,
            'time calc test 1',
        )
        entry2.resource = resource2

        # make sure an exception is NOT raised
        try:
            entry2.clean()
            self.assertTrue(entry1 == entry2)
        except Exception as e:
            traceback.print_exc()
            self.fail(
                'Cleaning an entry with \n'\
                'no time clash raised an unexpected exception: \n{0}'.format(e)
            )


    def test_entry_resource_clash_can_save_existing(self):
        """
        Make sure cleaning the data raises no Exception we are just posting back.
        """
        resource = create_resource('resource', 'resource')

        dateDelta1 = datetime.timedelta(days=0)
        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'time calc test 1',
        )
        entry1.resource = resource
        entry1.save()

        entry2 = Entry.objects.filter(
            date=entry1.date,
            time=entry1.time,
        ).first()
        # make sure an exception is NOT raised
        try:
            entry2.clean()
            self.assertTrue(entry1 == entry2)
        except Exception as e:
            traceback.print_exc()
            self.fail(
                'Cleaning an entry that already exists \n'\
                'raised an unexpected exception: \n{0}'.format(e)
            )


    def test_entry_resource_clash_no_resource_means_no_clash(self):
        """
        Entries using no resources do not generate any conflicts even when they
        overlap in time.
        """
        dateDelta1 = datetime.timedelta(days=0)
        time = datetime.time(hour=12)
        duration = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta1,
            time,
            duration,
            'resource_conflict_test_entry_1',
        )
        entry1.save()

        dateDelta2 = datetime.timedelta(days=0)
        entry2 = create_entry_from_delta(
            dateDelta2,
            time,
            duration,
            'resource_conflict_test_entry_2',
        )

        # make sure an exception is NOT raised
        try:
            entry2.clean()
            self.assertTrue(entry1 == entry2)
        except Exception as e:
            traceback.print_exc()
            self.fail(
                'Cleaning an entry with no resources \n'\
                'raised an unexpected exception: \n{0}'.format(e)
            )


    def test_entry_customer_double_booked(self):
        """
        Make sure customer cannot double-book (irrespective of resource).
        """
        customer = create_customer('test')

        dateDelta1 = datetime.timedelta(days=0)
        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'double-book test 1',
        )
        entry1.customer = customer
        entry1.save()

        entry2 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'double-book test 2',
        )
        entry2.customer = customer

        self.assertTrue(entry1 == entry2) # this passes the other tests
        self.assertRaisesMessage(
            ValidationError,
            'Double booking is not allowed. Please choose another time.',
            entry2.clean,
        )


    def test_cancelled_entry_customer_double_booked(self):
        """
        Make sure customer can double-book when one of the bookings is cancelled.
        """
        customer = create_customer('test')

        dateDelta1 = datetime.timedelta(days=0)
        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'double-book test 1',
        )
        entry1.customer = customer
        entry1.cancelled = True
        entry1.save()

        entry2 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'double-book test 2',
        )
        entry2.customer = customer

        # make sure an exception is NOT raised
        try:
            entry2.clean()
            self.assertTrue(entry1 == entry2)
        except Exception as e:
            traceback.print_exc()
            self.fail(
                'Cleaning a cancelled entry with double booking \n'\
                'raised an unexpected exception: \n{0}'.format(e)
            )


    @freeze_time('2015-10-12 12:00:00')
    def test_no_show_entry_customer_double_booked(self):
        """
        Make sure can double-book customer when one of the bookings is no-show.
        """
        customer = create_customer('test')

        dateDelta1 = datetime.timedelta(days=0)
        time1 = datetime.time(hour=12)
        duration1 = datetime.timedelta(hours=1)
        entry1 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'double-book test 1',
        )
        entry1.customer = customer
        entry1.no_show = True
        entry1.save()

        entry2 = create_entry_from_delta(
            dateDelta1,
            time1,
            duration1,
            'double-book test 2',
        )
        entry2.customer = customer

        # make sure an exception is NOT raised
        try:
            entry2.clean()
            self.assertTrue(entry1 == entry2)
        except Exception as e:
            traceback.print_exc()
            self.fail(
                'Cleaning a no_show entry with double booking \n'\
                'raised an unexpected exception: \n{0}'.format(e)
            )


    @freeze_time('2015-10-12 12:00:00')
    def test_entry_customer_out_of_hours(self):
        """
        Make sure customers cannot book outside trading hours.

        This test is patched to avoid complications with changing times.
        """

        customer = create_customer('test')

        dateDelta = datetime.timedelta(days=14)
        date = datetime.datetime.today().date()
        openingTime = settings.DIARY_OPENING_TIMES[date.weekday()]
        time1 = change_time(date, openingTime, datetime.timedelta(hours=1))
        duration = datetime.timedelta(hours=1)
        entry = create_entry_from_delta(
            dateDelta,
            time1,
            duration,
            'trading hours test 1',
        )
        entry.editor = customer

        # make sure an exception is NOT raised
        try:
            entry.clean()
            self.assertTrue(entry == entry)
        except Exception as e:
            traceback.print_exc()
            self.fail(
                'Cleaning an entry with no conflicts \n'\
                'raised an unexpected exception: \n{0}'.format(e)
            )

        # now try again out of hours
        time2 = change_time(date, openingTime, datetime.timedelta(hours=-1))
        entry.time = time2
        self.assertRaisesMessage(
            ValidationError,
            'Sorry, the store is closed then. Try changing the time.',
            entry.clean,
        )


    @freeze_time('2015-10-12 12:00:00')
    def test_entry_staff_out_of_hours(self):
        """
        Make sure staff are able to book any time.

        This test is patched to avoid complications with changing times.
        """

        dateDelta = datetime.timedelta(days=14)
        date = datetime.datetime.today().date()
        openingTime = settings.DIARY_OPENING_TIMES[date.weekday()]
        time1 = change_time(date, openingTime, datetime.timedelta(hours=-1))
        duration = datetime.timedelta(hours=1)
        entry = create_entry_from_delta(
            dateDelta,
            time1,
            duration,
            'trading hours test 2',
        )

        # make sure an exception is NOT raised
        try:
            entry.clean()
            self.assertTrue(entry == entry)
        except Exception as e:
            traceback.print_exc()
            self.fail(
                'Cleaning an entry with no conflicts \n'\
                'raised an unexpected exception: \n{0}'.format(e)
            )


    @freeze_time('2015-10-12 12:00:00')
    def test_entry_customer_in_past(self):
        """
        Make sure customers cannot book in the past.

        This test is patched to provide a constant value of 'now'
        """

        customer = create_customer('test')
        dateDelta = datetime.timedelta(days=0)
        date = timezone.localtime(timezone.now()).date()
        now = timezone.localtime(timezone.now()).time()
        time = change_time(date, now, datetime.timedelta(hours=-1))
        duration = datetime.timedelta(hours=1)
        entry = create_entry_from_delta(
            dateDelta,
            time,
            duration,
            'past test 1',
        )
        entry.editor = customer

        self.assertRaisesMessage(
            ValidationError,
            'Please book a date/time in the future.',
            entry.clean,
        )


    @freeze_time('2015-10-12 12:00:00')
    def test_entry_customer_advance_limit(self):
        """
        Make sure customers cannot book before the advance limit. (Time in
        future but before the advance booking limit).

        This test is patched to provide a constant value of 'now'
        """

        # set up the min_booking time
        min_booking_orig = settings.DIARY_MIN_BOOKING
        setattr(main_settings, 'DIARY_MIN_BOOKING', 1)
        imp.reload(settings)

        customer = create_customer('test')
        dateDelta = datetime.timedelta(days=0)
        date = timezone.localtime(timezone.now()).date()
        now = timezone.localtime(timezone.now()).time()
        time = change_time(date, now, datetime.timedelta(hours=1))
        duration = datetime.timedelta(hours=1)
        entry = create_entry_from_delta(
            dateDelta,
            time,
            duration,
            'advance test',
        )
        entry.editor = customer

        self.assertRaisesMessage(
            ValidationError,
            'Need to book ahead.',
            entry.clean,
        )

        # reset the settings
        setattr(main_settings, 'DIARY_MIN_BOOKING', min_booking_orig)
        imp.reload(settings)



class ViewTests(TestCase):
    """
    Tests of various aspects of the diary views and behavior.
    """


    def setup(self):
        """
        Set up a client for getting/posting the views.
        """
        User.objects.create_user('test', 'test@example.com', 'test')
        self.client.login(username='test', password='test')


    def test_login_required_for_diary(self):
        """
        Navigating to the diary automatically redirects to login page if
        not logged in.
        """
        response = self.client.get(reverse('diary:year_now'), follow=True)
        self.assertRedirects(
            response,
            '/accounts/login/?next=/diary/year/',
        )


    def test_cal_first_day_of_week_default(self):
        """
        Make sure the custom setting for day of week works.
        """
        self.setup()

        # test the default
        self.assertEqual(views.DAY_NAMES[0], 'Monday')


    def test_cal_first_day_of_week_sunday(self):
        """
        Make sure the custom setting for day of week works.
        """
        self.setup()
        first_day_of_week_orig = settings.DIARY_FIRST_DAY_OF_WEEK
        setattr(main_settings, 'DIARY_FIRST_DAY_OF_WEEK', 6)

        # test for sunday
        imp.reload(settings)
        imp.reload(views)
        self.assertEqual(views.DAY_NAMES[0], 'Sunday')

        # tidy up the mess (make sure default is restored)
        setattr(main_settings, 'DIARY_FIRST_DAY_OF_WEEK', first_day_of_week_orig)
        imp.reload(settings)
        imp.reload(views)
        self.assertEqual(views.DAY_NAMES[0], 'Monday')


def create_entries():
    tenYearsAgo = yearsago(10)
    fiveYearsAgo = yearsago(5)
    now = yearsago(0)
    time = datetime.time(hour=12)
    duration = datetime.timedelta(hours=1)
    entry1 = create_entry(tenYearsAgo, time, duration, "very old entry")
    entry1.save()
    entry2 = create_entry(fiveYearsAgo, time, duration, "old entry")
    entry2.save()
    entry3 = create_entry(now, time, duration, "new")
    entry3.save()


class Clean_Entry_Tests(TestCase):
    """
    Tests of the management command to clear out old entries.
    """

    @freeze_time('2020-07-01 12:00:00')
    def test_no_args_gives_error(self):
        """
        Make sure providing no arguments raises an error.
        """
        msg = "Specify a valid before date (-b) or an age (-a)."
        with self.assertRaisesMessage(CommandError, msg):
            call_command('clean_entries')


    @freeze_time('2020-07-01 12:00:00')
    def test_negative_age_gives_error(self):
        """
        Make sure negative age raises an error.
        """
        msg = "Specify a valid before date (-b) or an age (-a)."
        with self.assertRaisesMessage(CommandError, msg):
            call_command('clean_entries', '-a=-3')


    @freeze_time('2020-07-01 12:00:00')
    def test_future_date_gives_error(self):
        """
        Make sure future date raises an error.
        """
        today, now = get_today_now()
        futureDate = today.replace(year=today.year + 5)
        futureDateString = futureDate.isoformat()
        msg = "Specify a valid before date (-b) or an age (-a)."
        with self.assertRaisesMessage(CommandError, msg):
            call_command('clean_entries', '-b=%s' % (futureDateString))


    @freeze_time('2020-07-01 12:00:00')
    def test_very_old_entry_delete_by_age(self):
        """
        Make sure very old data is deleted by age.
        """
        create_entries()
        call_command('clean_entries', '-a=8')
        entries = Entry.objects.all()
        self.assertEqual(2, len(entries))


    @freeze_time('2020-07-01 12:00:00')
    def test_very_old_entry_delete_by_before(self):
        """
        Make sure very old data is deleted by before date.
        """
        create_entries()
        call_command('clean_entries', '-b=2012-07-01')
        entries = Entry.objects.all()
        self.assertEqual(2, len(entries))


    @freeze_time('2020-07-01 12:00:00')
    def test_age_overrides_before(self):
        """
        Make sure age criteria overrides the given before date.
        """
        create_entries()
        call_command('clean_entries', '-a=3', '-b=2012-07-01')
        entries = Entry.objects.all()
        print("Entries after clean=", len(entries))
        self.assertEqual(1, len(entries))



from django.core.management.base import BaseCommand, CommandError
from diary.models import Entry, Customer
from diary.views import get_today_now
import datetime


class Command(BaseCommand):
    """
    Periodically clean out old database data in the Entry table.
    """

    help = "Clean old Entry data in the database."


    def add_arguments(self, parser):
        """
        Define the date before when to delete entries either by
        -a --age      age of the entry (from the date of the appointment)
        -b --before   date before which to delete
        """
        parser.add_argument(
            '-a',
            '--age',
            help="age in years",
            type=int,
            default=0,
        )
        parser.add_argument(
            '-b',
            '--before',
            help="date HWM (yyyy-mm-dd)",
            type=datetime.date.fromisoformat,
            default=get_today_now()[0],
        )


    def handle(self, *args, **kwargs):
        """
        Implement the database cleanout using the argument(s) provided.
        """

        today, now = get_today_now()

        age = kwargs['age']
        before = kwargs['before']

        # override before date with date calculated from age
        if age > 0:
            before = today.replace(year=today.year-age)

        # ensure the date being asked for is sane
        if not before or before >= today:
            raise CommandError("Specify a valid before date (-b) or an age (-a).")

        # delete entries before the selected date
        print("Cleaning entries prior to ", before)
        Entry.objects.filter(date__lt=before,).delete()



from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import User
from .models import Customer


class CustomerAuthBackend(ModelBackend):
    """
    Home-spun backend in response to django-model-utils functionality not
    working for Django >~ 2
    """
    def get_user(self, user_id):
        try:
            return Customer.get(pk=user_id)
        except User.DoesNotExist:
            return None

    def authenticate(self, username=None, password=None):
        try:
            user = Customer.get(username=username)
            if user.check_password(password):
                return user
        except User.DoesNotExist:
            return None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


def add_editor(apps, schema_editor):
    """
    Add an editor to existing entries.
    
    We choose the customer if it exists, or the creator if it does not.
    """
    Entry = apps.get_model('diary', 'Entry')
    for entry in Entry.objects.all():
        entry.editor = entry.customer if entry.customer else entry.creator
        entry.save()



class Migration(migrations.Migration):

    dependencies = [
        ('diary', '0002_auto_20151010_1324'),
    ]

    operations = [
        migrations.RunPython(add_editor),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.core.validators
import datetime
import diary.models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('auth', '0006_require_contenttypes_0002'),
    ]

    operations = [
        migrations.CreateModel(
            name='Customer',
            fields=[
                ('user_ptr', models.OneToOneField(auto_created=True, parent_link=True, serialize=False, to=settings.AUTH_USER_MODEL, primary_key=True, on_delete=models.CASCADE)),
                ('phone', models.CharField(max_length=20, blank=True, validators=[django.core.validators.RegexValidator(message='Not a valid phone number', regex='[0-9][0-9 ]+')], null=True)),
                ('date_of_birth', models.DateField(blank=True, null=True)),
                ('gender', models.CharField(max_length=1, default='F', choices=[('M', 'Male'), ('F', 'Female')])),
                ('notes', models.TextField(blank=True)),
            ],
            options={
                'verbose_name': 'Customer',
                'verbose_name_plural': 'Customers',
            },
            bases=('auth.user',),
            managers=[
                ('objects', diary.models.CustomerManager()),
            ],
        ),
        migrations.CreateModel(
            name='Entry',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, verbose_name='ID', primary_key=True)),
                ('date', models.DateField()),
                ('time', models.TimeField(default=datetime.time(12, 0))),
                ('duration', models.TimeField(blank=True, default=datetime.time(1, 0))),
                ('notes', models.TextField(blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('creator', models.ForeignKey(blank=True, null=True, related_name='created_entries', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE),),
                ('customer', models.ForeignKey(blank=True, null=True, related_name='entries', to='diary.Customer', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'entries',
            },
        ),
        migrations.CreateModel(
            name='Resource',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, verbose_name='ID', primary_key=True)),
                ('name', models.CharField(max_length=40)),
                ('description', models.TextField(blank=True)),
            ],
        ),
        migrations.CreateModel(
            name='Treatment',
            fields=[
                ('id', models.AutoField(auto_created=True, serialize=False, verbose_name='ID', primary_key=True)),
                ('name', models.CharField(max_length=40)),
                ('min_duration', models.DurationField(blank=True)),
                ('resource_required', models.BooleanField(default=False)),
            ],
        ),
        migrations.AddField(
            model_name='entry',
            name='resource',
            field=models.ForeignKey(blank=True, null=True, to='diary.Resource', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='entry',
            name='treatment',
            field=models.ForeignKey(blank=True, null=True, to='diary.Treatment', on_delete=models.CASCADE),
        ),
    ]



from django.urls import reverse
from django.utils.safestring import mark_safe
from django.forms import widgets
from django.conf import settings
from django.utils.translation import ugettext as _




class RelatedFieldWidgetCanAdd(widgets.Select):
    """
    Supplements the standard Select widget behaviour by adding an 'add' button.

    Added related_kwargs for targeted redirection.

    Adapted from:
    http://stackoverflow.com/questions/28068168/django-adding-an-add-new-button-for-a-foreignkey-in-a-modelform
    """


    def __init__(
        self,
        related_model,
        related_url=None,
        related_kwargs=None, *args, **kwargs
    ):

        super(RelatedFieldWidgetCanAdd, self).__init__(*args, **kwargs)

        if not related_url:         # construct a url for the default admin site
            rel_to = related_model
            info = (rel_to._meta.app_label, rel_to._meta.object_name.lower())
            related_url = 'admin:%s_%s_add' % info
            related_kwargs = {}

        # Be careful that here "reverse" is not allowed
        self.related_url = related_url
        self.related_kwargs = related_kwargs


    def render(self, name, value, *args, **kwargs):
        self.related_url = reverse(
            self.related_url,
            kwargs=self.related_kwargs,
        )
        output = [
            super(RelatedFieldWidgetCanAdd, self).render(
                name,
                value,
                *args,
                **kwargs,
            )
        ]
        output.append('<a href="%s" class="add-another" id="add_id_%s" onclick="return showAddAnotherPopup(this);"> ' % \
            (self.related_url, name))
        output.append('<img src="%sadmin/img/icon-addlink.svg" width="14" height="14" alt="%s"/></a>' % (settings.STATIC_URL, _('+ New')))
        return mark_safe(''.join(output))
